@@ -150,8 +150,42 @@ class DataIntegrityError(RuntimeError):
     Data integrity seems to have been lost.
     """
 
+
+
 class BrokenReference(DataIntegrityError):
     """
     A reference to a nonexistent item was detected when this should be
     impossible.
     """
+
+
+
+class UpgraderRecursion(RuntimeError):
+    """
+    Upgraders are not allowed to recurse.
+    """
+
+
+
+class ItemUpgradeError(RuntimeError):
+    """
+    Attempting to upgrade an Item resulted in an error.
+
+    @ivar storeID: Store ID of the item that failed to upgrade
+    @ivar oldType: The type of the item being upgraded
+    @ivar newType: The type the item should've been upgraded to
+    """
+    def __init__(self, storeID, oldType, newType):
+        RuntimeError.__init__(self, storeID, oldType, newType)
+        self.storeID = storeID
+        self.oldType = oldType
+        self.newType = newType
+
+
+
+class UnsatisfiedRequirement(AttributeError):
+    """
+    A requirement described by a L{axiom.dependency.requiresFromSite} was not
+    satisfied by the database, and could not be satisfied automatically at
+    runtime by a default factory.
+    """
