@@ -2,9 +2,11 @@
 
 __metaclass__ = type
 
-from zope.interface import implements
+import gc
+from zope.interface import implements, Interface
 
 from inspect import getabsfile
+from weakref import WeakValueDictionary
 
 from twisted.python import log
 from twisted.python.reflect import qual, namedAny
@@ -21,7 +23,7 @@ from axiom.attributes import (
     _OrderingMixin, _ContainableMixin, Comparable, compare, inmemory,
     reference, text, integer, AND, _cascadingDeletes, _disallows)
 
-_typeNameToMostRecentClass = {}
+_typeNameToMostRecentClass = WeakValueDictionary()
 
 def normalize(qualName):
     """
@@ -60,6 +62,9 @@ class MetaItem(slotmachine.SchemaMetaMachine):
             T.typeName = normalize(qual(T))
         if T.schemaVersion is None:
             T.schemaVersion = 1
+        if T.typeName in _typeNameToMostRecentClass:
+            # Let's try not to gc.collect() every time.
+            gc.collect()
         if T.typeName in _typeNameToMostRecentClass:
             if T.__legacy__:
                 return T
@@ -74,8 +79,10 @@ class MetaItem(slotmachine.SchemaMetaMachine):
                 else:
                     relmod = otherT.__module__
 
-                raise RuntimeError("Use absolute imports; relative import"
-                        " detected for type %r (imported from %r)" % (T.typeName, relmod))
+                raise RuntimeError(
+                    "Use absolute imports; relative import"
+                    " detected for type %r (imported from %r)" % (
+                        T.typeName, relmod))
 
             raise RuntimeError("2 definitions of axiom typename %r: %r %r" % (
                     T.typeName, T, _typeNameToMostRecentClass[T.typeName]))
@@ -90,7 +97,8 @@ class MetaItem(slotmachine.SchemaMetaMachine):
         purposes.
         """
         if isinstance(other, MetaItem):
-            return cmp(self.typeName, other.typeName)
+            return cmp((self.typeName, self.schemaVersion),
+                       (other.typeName, other.schemaVersion))
         return NotImplemented
 
 
@@ -170,12 +178,40 @@ def serviceSpecialCase(item, pups):
 
 
 class Empowered(object):
+    """
+    An object which can have powerups.
+
+    @type store: L{axiom.store.Store}
+    @ivar store: Persistence object to which powerups can be added for later
+        retrieval.
+
+    @type aggregateInterfaces: C{dict}
+    @ivar aggregateInterfaces: Mapping from interface classes to callables
+        which will be used to produce corresponding powerups.  The callables
+        will be invoked with two arguments, the L{Empowered} for which powerups
+        are being loaded and with a list of powerups found in C{store}.  The
+        return value is the powerup.  These are used only by the callable
+        interface adaption API, not C{powerupsFor}.
+    """
 
     aggregateInterfaces = {
         IService: serviceSpecialCase,
         IServiceCollection: serviceSpecialCase}
 
-    def powerUp(self, powerup, interface, priority=0):
+    def inMemoryPowerUp(self, powerup, interface):
+        """
+        Install an arbitrary object as a powerup on an item or store.
+
+        Powerups installed using this method will only exist as long as this
+        object remains in memory.  They will also take precedence over powerups
+        installed with L{powerUp}.
+
+        @param interface: a zope interface
+        """
+        self._inMemoryPowerups[interface] = powerup
+
+
+    def powerUp(self, powerup, interface=None, priority=0):
         """
         Installs a powerup (e.g. plugin) on an item or store.
 
@@ -194,8 +230,23 @@ class Empowered(object):
         30, the powerup will be adjusted to priority 30 but future calls to
         powerupFor will still only return that powerup once.
 
-        @param powerup: an Item that implements C{interface}
-        @param interface: a zope interface
+
+        If no interface or priority are specified, and the class of the
+        powerup has a "powerupInterfaces" attribute (containing
+        either a sequence of interfaces, or a sequence of
+        (interface, priority) tuples), this object will be powered up
+        with the powerup object on those interfaces.
+
+        If no interface or priority are specified and the powerup has
+        a "__getPowerupInterfaces__" method, it will be called with
+        an iterable of (interface, priority) tuples, collected from the
+        "powerupInterfaces" attribute described above. The iterable of
+        (interface, priority) tuples it returns will then be
+        installed.
+
+
+        @param powerup: an Item that implements C{interface} (if specified)
+        @param interface: a zope interface, or None
 
         @param priority: An int; preferably either POWERUP_BEFORE,
         POWERUP_AFTER, or unspecified.
@@ -204,29 +255,53 @@ class Empowered(object):
         install a powerup for IPowerupIndirector because that would be
         nonsensical.
         """
-        if interface is IPowerupIndirector:
+        if interface is None:
+            for iface, priority in powerup._getPowerupInterfaces():
+                self.powerUp(powerup, iface, priority)
+
+        elif interface is IPowerupIndirector:
             raise TypeError(
                 "You cannot install a powerup for IPowerupIndirector: " +
                 powerup)
-
-        forc = self.store.findOrCreate(_PowerupConnector,
-                                       item=self,
-                                       interface=unicode(qual(interface)),
-                                       powerup=powerup)
-        forc.priority = priority
+        else:
+            forc = self.store.findOrCreate(_PowerupConnector,
+                                           item=self,
+                                           interface=unicode(qual(interface)),
+                                           powerup=powerup)
+            forc.priority = priority
 
 
-    def powerDown(self, powerup, interface):
+    def powerDown(self, powerup, interface=None):
         """
         Remove a powerup.
+
+        If no interface is specified, and the type of the object being
+        installed has a "powerupInterfaces" attribute (containing
+        either a sequence of interfaces, or a sequence of (interface,
+        priority) tuples), the target will be powered down with this
+        object on those interfaces.
+
+        If this object has a "__getPowerupInterfaces__" method, it
+        will be called with an iterable of (interface, priority)
+        tuples. The iterable of (interface, priority) tuples it
+        returns will then be uninstalled.
+
+        (Note particularly that if powerups are added or removed to the
+        collection described above between calls to powerUp and powerDown, more
+        powerups or less will be removed than were installed.)
         """
-        for cable in self.store.query(_PowerupConnector,
-                                      AND(_PowerupConnector.item == self,
-                                          _PowerupConnector.interface == unicode(qual(interface)),
-                                          _PowerupConnector.powerup == powerup)):
-            cable.deleteFromStore()
-            return
-        raise ValueError("Not powered up for %r with %r" % (interface, powerup))
+        if interface is None:
+            for interface, priority in powerup._getPowerupInterfaces():
+                self.powerDown(powerup, interface)
+        else:
+            for cable in self.store.query(_PowerupConnector,
+                                          AND(_PowerupConnector.item == self,
+                                              _PowerupConnector.interface == unicode(qual(interface)),
+                                              _PowerupConnector.powerup == powerup)):
+                cable.deleteFromStore()
+                return
+            raise ValueError("Not powered up for %r with %r" % (interface,
+                                                                powerup))
 
 
     def __conform__(self, interface):
@@ -253,6 +328,7 @@ class Empowered(object):
         for p in pups:
             return p
 
+
     def powerupsFor(self, interface):
         """
         Returns powerups installed using C{powerUp}, in order of descending
@@ -262,6 +338,9 @@ class Empowered(object):
         powerupsFor iteration, during an upgrader, or previously, will not be
         returned.
         """
+        inMemoryPowerup = self._inMemoryPowerups.get(interface, None)
+        if inMemoryPowerup is not None:
+            yield inMemoryPowerup
         name = unicode(qual(interface), 'ascii')
         for cable in self.store.query(
             _PowerupConnector,
@@ -280,6 +359,13 @@ class Empowered(object):
                     yield pup
 
     def interfacesFor(self, powerup):
+        """
+        Return an iterator of the interfaces for which the given powerup is
+        installed on this object.
+
+        This is not implemented for in-memory powerups.  It will probably fail
+        in an unpredictable, implementation-dependent way if used on one.
+        """
         pc = _PowerupConnector
         for iface in self.store.query(pc,
                                       AND(pc.item == self,
@@ -287,6 +373,32 @@ class Empowered(object):
             yield namedAny(iface)
 
 
+    def _getPowerupInterfaces(self):
+        """
+        Collect powerup interfaces this object declares that it can be
+        installed on.
+        """
+        powerupInterfaces = getattr(self.__class__, "powerupInterfaces", ())
+        pifs = []
+        for x in powerupInterfaces:
+            if isinstance(x, type(Interface)):
+                #just an interface
+                pifs.append((x, 0))
+            else:
+                #an interface and a priority
+                pifs.append(x)
+
+        m = getattr(self, "__getPowerupInterfaces__", None)
+        if m is not None:
+            pifs = m(pifs)
+            try:
+                pifs = [(i, p) for (i, p) in pifs]
+            except ValueError:
+                raise ValueError("return value from %r.__getPowerupInterfaces__"
+                                 " not an iterable of 2-tuples" % (self,))
+        return pifs
+
+
 def transacted(func):
     """
     Return a callable which will invoke C{func} in a transaction using the
@@ -381,6 +493,9 @@ class Item(Empowered, slotmachine._Strict):
     _storeIDComparer = None
     _axiom_service = inmemory()
 
+    # A mapping from interfaces to in-memory powerups.
+    _inMemoryPowerups = inmemory()
+
     def _currentlyValidAsReferentFor(self, store):
         """
         Is this object currently valid as a reference?  Objects which will be
@@ -424,7 +539,8 @@ class Item(Empowered, slotmachine._Strict):
             self.__store = store
             oid = self.storeID = self.store.executeSchemaSQL(
                 _schema.CREATE_OBJECT, [self.store.getTypeID(type(self))])
-            store.objectCache.cache(oid, self)
+            if not self.__legacy__:
+                store.objectCache.cache(oid, self)
             if store.autocommit:
                 log.msg(interface=iaxiom.IStatEvent,
                         name='database', stat_autocommits=1)
@@ -470,6 +586,7 @@ class Item(Empowered, slotmachine._Strict):
         instantiation or loading from the database.
         """
         self._axiom_service = None
+        self._inMemoryPowerups = {}
         self.__dirty__ = {}
         to__store = kw.pop('__store', None)
         to__everInserted = kw.pop('__everInserted', False)
@@ -498,11 +615,18 @@ class Item(Empowered, slotmachine._Strict):
 
                 self._schemaPrepareInsert(tostore)
                 self.__store = tostore
+
+                # However, setting the store would normally cache this item as
+                # well, so we need to cache it here - unless this is actually a
+                # dummy class which isn't real!  In that case don't.
+                if not self.__legacy__:
+                    tostore.objectCache.cache(self.storeID, self)
                 if tostore.autocommit:
                     self.checkpoint()
             else:
                 self.store = tostore
 
+
     def __init__(self, **kw):
         """
         Create a new Item.  This is called on an item *only* when it is being created
@@ -521,9 +645,11 @@ class Item(Empowered, slotmachine._Strict):
         self.__justCreated = True
         self.__subinit__(**kw)
 
+
     def __finalizer__(self):
         return noop
 
+
     def existingInStore(cls, store, storeID, attrs):
         """Create and return a new instance from a row from the store."""
         self = cls.__new__(cls)
@@ -722,11 +848,6 @@ class Item(Empowered, slotmachine._Strict):
         new.touch()
         new.activate()
 
-        # XXX TODO; this needs to be forced to fall out of cache in the case of
-        # an in memory revert
-        if not new.__legacy__:
-            self.store.objectCache.cache(self.storeID, new)
-
         self.store.executeSchemaSQL(_schema.CHANGE_TYPE,
                                     [newTypeID, self.storeID])
         self.deleteFromStore(False)
