@@ -10,7 +10,7 @@ from twisted.python import log, failure
 from epsilon.extime import Time
 from epsilon import descriptor
 
-from axiom.item import Item
+from axiom.item import Item, declareLegacyItem
 from axiom.attributes import AND, timestamp, reference, integer, inmemory, bytes
 from axiom.dependency import dependsOn, installOn
 from axiom.iaxiom import IScheduler
@@ -36,6 +36,12 @@ class TimedEvent(Item):
     time = timestamp(indexed=True)
     runnable = reference()
 
+    running = inmemory(doc='True if this event is currently running.')
+
+    def activate(self):
+        self.running = False
+
+
     def _rescheduleFromRun(self, newTime):
         """
         Schedule this event to be run at the indicated time, or if the
@@ -56,7 +62,12 @@ class TimedEvent(Item):
         if runnable is None:
             self.deleteFromStore()
         else:
-            self._rescheduleFromRun(runnable.run())
+            try:
+                self.running = True
+                newTime = runnable.run()
+            finally:
+                self.running = False
+            self._rescheduleFromRun(newTime)
 
 
     def handleError(self, now, failureObj):
@@ -76,11 +87,11 @@ class TimedEvent(Item):
 
 
     def _defaultErrorHandler(self, now, failureObj):
-        tefl = TimedEventFailureLog(store=self.store,
-                                    desiredTime=self.time,
-                                    actualTime=now,
-                                    runnable=self.runnable,
-                                    traceback=failureObj.getTraceback())
+        TimedEventFailureLog(store=self.store,
+                             desiredTime=self.time,
+                             actualTime=now,
+                             runnable=self.runnable,
+                             traceback=failureObj.getTraceback())
         self.deleteFromStore()
 
 
@@ -95,11 +106,6 @@ class _WackyControlFlow(Exception):
 MAX_WORK_PER_TICK = 10
 
 class SchedulerMixin:
-    def now(self):
-        # Testing hook
-        return Time()
-
-
     def _oneTick(self, now):
         theEvent = self._getNextEvent(now)
         if theEvent is None:
@@ -115,7 +121,7 @@ class SchedulerMixin:
     def _getNextEvent(self, now):
         # o/` gonna party like it's 1984 o/`
         theEventL = list(self.store.query(TimedEvent,
-                                          TimedEvent.time < now,
+                                          TimedEvent.time <= now,
                                           sort=TimedEvent.time.ascending,
                                           limit=1))
         if theEventL:
@@ -125,7 +131,6 @@ class SchedulerMixin:
     def tick(self):
         now = self.now()
         self.nextEventAt = None
-        before = self.eventsRun
         workBeingDone = True
         workUnitsPerformed = 0
         errors = 0
@@ -188,7 +193,7 @@ class SchedulerMixin:
         """
         events = self.store.query(
             TimedEvent, TimedEvent.runnable == runnable)
-        return events.getColumn("time")
+        return (event.time for event in events if not event.running)
 
 _EPSILON = 1e-20      # A very small amount of time.
 
@@ -211,9 +216,10 @@ class Scheduler(Item, Service, SchedulerMixin):
     callLater = inmemory()
     now = inmemory()
 
-    eventsRun = integer()
-    lastEventAt = timestamp()
-    nextEventAt = timestamp()
+    # This is unused and should be upgraded away.
+    eventsRun = integer(default=0)
+    lastEventAt = timestamp(default=None)
+    nextEventAt = timestamp(default=None)
 
     class running(descriptor.attribute):
         def get(self):
@@ -227,13 +233,6 @@ class Scheduler(Item, Service, SchedulerMixin):
             pass
 
 
-    def __init__(self, **kw):
-        super(Scheduler, self).__init__(**kw)
-        self.eventsRun = 0
-        self.lastEventAt = None
-        self.nextEventAt = None
-
-
     def __repr__(self):
         return '<Scheduler>'
 
@@ -289,27 +288,32 @@ class Scheduler(Item, Service, SchedulerMixin):
         self.nextEventAt = when
 
 
+
 class _SubSchedulerParentHook(Item):
-    schemaVersion = 2
+    schemaVersion = 3
     typeName = 'axiom_subscheduler_parent_hook'
 
     loginAccount = reference()
-    scheduledAt = timestamp(default=None)
 
     scheduler = dependsOn(Scheduler)
 
     def run(self):
-        self.scheduledAt = None
+        """
+        Tick our C{loginAccount}'s L{SubScheduler}.
+        """
         IScheduler(self.loginAccount).tick()
 
+
     def _schedule(self, when):
-        if self.scheduledAt is not None:
-            if when < self.scheduledAt:
-                self.scheduler.reschedule(self, self.scheduledAt, when)
-                self.scheduledAt = when
+        """
+        Ensure that this hook is scheduled to run at or before C{when}.
+        """
+        for scheduledAt in self.scheduler.scheduledTimes(self):
+            if when < scheduledAt:
+                self.scheduler.reschedule(self, scheduledAt, when)
+            break
         else:
             self.scheduler.schedule(self, when)
-            self.scheduledAt = when
 
 
 def upgradeParentHook1to2(oldHook):
@@ -325,6 +329,24 @@ def upgradeParentHook1to2(oldHook):
 
 registerUpgrader(upgradeParentHook1to2, _SubSchedulerParentHook.typeName, 1, 2)
 
+declareLegacyItem(
+    _SubSchedulerParentHook.typeName, 2,
+    dict(loginAccount=reference(),
+         scheduledAt=timestamp(default=None),
+         scheduler=reference()))
+
+def upgradeParentHook2to3(old):
+    """
+    Copy all attributes except C{scheduledAt}.
+    """
+    return old.upgradeVersion(
+        old.typeName, 2, 3,
+        loginAccount=old.loginAccount,
+        scheduler=old.scheduler)
+
+registerUpgrader(upgradeParentHook2to3, _SubSchedulerParentHook.typeName, 2, 3)
+
+
 
 class SubScheduler(Item, SchedulerMixin):
     """
@@ -342,6 +364,7 @@ class SubScheduler(Item, SchedulerMixin):
     nextEventAt = timestamp()
 
     # Also testing hooks
+    callLater = inmemory()
     now = inmemory()
 
     def __repr__(self):
@@ -349,6 +372,7 @@ class SubScheduler(Item, SchedulerMixin):
 
 
     def activate(self):
+        self.callLater = reactor.callLater
         self.now = Time
 
     def _transientSchedule(self, when, now):
@@ -382,4 +406,4 @@ class SubScheduler(Item, SchedulerMixin):
         """
         te = self.store.findFirst(TimedEvent, sort=TimedEvent.time.descending)
         if te is not None:
-            self._transientSchedule(te.time, Time)
+            self._transientSchedule(te.time, self.now)
