@@ -1,3 +1,4 @@
+# Copyright 2008 Divmod, Inc.  See LICENSE for details
 # -*- test-case-name: axiom.test -*-
 
 """
@@ -7,12 +8,7 @@ This module holds the Axiom Store class and related classes, such as queries.
 from epsilon import hotfix
 hotfix.require('twisted', 'filepath_copyTo')
 
-import time
-import os
-import itertools
-import warnings
-import sys
-import operator
+import time, os, itertools, warnings, sys, operator, weakref
 
 from zope.interface import implements
 
@@ -20,7 +16,7 @@ from twisted.python import log
 from twisted.python.failure import Failure
 from twisted.python import filepath
 from twisted.internet import defer
-from twisted.python.reflect import namedAny, qual
+from twisted.python.reflect import namedAny
 from twisted.python.util import unsignedID
 from twisted.application.service import IService, IServiceCollection
 
@@ -29,25 +25,32 @@ from epsilon.cooperator import SchedulingService
 
 from axiom import _schema, attributes, upgrade, _fincache, iaxiom, errors, batch
 from axiom import item
-
-# Doing this in a slightly awkward way so Pyflakes won't complain; it really
-# doesn't like conditional imports.
-if attributes.USING_APSW:
-    backendName = 'axiom._apsw.Connection'
-else:
-    backendName = 'axiom._pysqlite2.Connection'
-
-Connection = namedAny(backendName)
-
+from axiom._pysqlite2 import Connection
 
 from axiom.item import \
-    _typeNameToMostRecentClass, declareLegacyItem,\
+    _typeNameToMostRecentClass, declareLegacyItem, \
     _legacyTypes, Empowered, serviceSpecialCase, _StoreIDComparer
 
 IN_MEMORY_DATABASE = ':memory:'
 
+# The special storeID used to mark the store itself as the target of a
+# reference.
+STORE_SELF_ID = -1
+
 tempCounter = itertools.count()
 
+# A mapping from MetaItem instances to precomputed structures describing the
+# indexes necessary for those MetaItems.  Avoiding recomputing this speeds up
+# opening stores significantly.
+_requiredTableIndexes = weakref.WeakKeyDictionary()
+
+# A mapping from MetaItem instances to precomputed structures describing the
+# known in-memory schema for those MetaItems.  Avoiding recomputing this speeds
+# up opening stores significantly.
+_inMemorySchemaCache = weakref.WeakKeyDictionary()
+
+
+
 class NoEmptyItems(Exception):
     """You must define some attributes on every item.
     """
@@ -281,15 +284,6 @@ class BaseQuery:
 
 
     def _sqlAndArgs(self, verb, subject):
-
-        # Generate the WHERE clause separately from determining the tables
-        # which are involved so that the loop over those tables above has a
-        # chance to call getTableAlias, which may have side-effects.
-        if self.comparison is not None:
-            where = 'WHERE ' + self.comparison.getQuery(self.store)
-        else:
-            where = ''
-
         limitClause = []
         if self.limit is not None:
             # XXX LIMIT and OFFSET used to be using ?, but they started
@@ -972,6 +966,15 @@ class Store(Empowered):
     stored.  This cache is generated from the saved schema metadata when this
     store is opened and updated when schema changes from other store objects
     (such as in other processes) are detected.
+
+    @cvar __legacy__: an L{Item} may refer to a L{Store} via a L{reference},
+    and this attribute tells the item reference system that the store itself is
+    not an old version of an item; i.e. it does not need to have its upgraders
+    invoked.
+
+    @cvar storeID: an L{Item} may refer to a L{Store} via a L{reference}, and
+    this attribute tells the item reference system that the L{Store} has a
+    special ID that to use (which is never allocated to any item).
     """
 
     aggregateInterfaces = {
@@ -984,9 +987,6 @@ class Store(Empowered):
     transaction = None          # set of objects changed in the current transaction
     touched = None              # set of objects changed since the last checkpoint
 
-    storeID = -1                # I have a StoreID so that things can reference
-                                # me
-
     databaseName = 'main'       # can differ if database is attached to another
                                 # database.
 
@@ -1005,22 +1005,43 @@ class Store(Empowered):
     # non-zero will reject database changes with a ChangeRejected exception.
     _rejectChanges = 0
 
+    # The following method and attributes are the ad-hoc interface required as
+    # targets of attributes.reference attributes.  (In other words, the store
+    # is a little bit like a fake item.)  These should probably eventually be
+    # on an interface somewhere, and be better named.
+
     def _currentlyValidAsReferentFor(self, store):
-        """necessary because I can be a target of attributes.reference()
+        """
+        Check to see if this store is currently valid as a target of a
+        reference from an item in the given L{Store}.  This is true iff the
+        given L{Store} is this L{Store}.
+
+        @param store: the store that the referring item is present in.
+
+        @type store: L{Store}
         """
         if store is self:
             return True
         else:
             return False
 
-    def __init__(self, dbdir=None, debug=False, parent=None, idInParent=None):
+    __legacy__ = False
+
+    storeID = STORE_SELF_ID
+
+
+    def __init__(self, dbdir=None, filesdir=None, debug=False, parent=None, idInParent=None):
         """
         Create a store.
 
-        @param dbdir: A name of an existing Axiom directory, or the name of a
+        @param dbdir: A L{FilePath} to (or name of) an existing Axiom directory, or
         directory that does not exist yet which will be created as this Store
         is instantiated.  If unspecified, this database will be kept in memory.
 
+        @param filesdir: A L{FilePath} to (or name of) a directory to keep files in for in-memory
+        stores. An exception will be raised if both this attribute and C{dbdir}
+        are specified.
+
         @param debug: set to True if this Store should print out every SQL
         statement it sends to SQLite.
 
@@ -1030,6 +1051,8 @@ class Store(Empowered):
         @param idInParent: (internal) If this is opened using an
         L{axiom.substore.Substore}, the storeID of the item within its parent
         which opened it.
+
+        @raises: C{ValueError} if both C{dbdir} and C{filesdir} are specified.
         """
         if parent is not None or idInParent is not None:
             assert parent is not None
@@ -1041,6 +1064,8 @@ class Store(Empowered):
         self.queryTimes = []
         self.execTimes = []
 
+        self._inMemoryPowerups = {}
+
         self._attachedChildren = {} # database name => child store object
 
         self.statementCache = {} # non-normalized => normalized qmark SQL
@@ -1066,8 +1091,7 @@ class Store(Empowered):
         self.typeToTableNameCache = {}
         self.attrToColumnNameCache = {}
 
-        self._oldTypesRemaining = [] # a list of old types which have not been
-                                     # fully upgraded in this database.
+        self._upgradeManager = upgrade._StoreUpgrade(self)
 
         self._axiom_service = None
 
@@ -1087,7 +1111,18 @@ class Store(Empowered):
         if dbdir is None:
             self._initdb(IN_MEMORY_DATABASE)
             self._initSchema()
+            self._memorySubstores = []
+            if filesdir is not None:
+                if not isinstance(filesdir, filepath.FilePath):
+                    filesdir = filepath.FilePath(filesdir)
+                self.filesdir = filesdir
+                if not self.filesdir.isdir():
+                    self.filesdir.makedirs()
+                    self.filesdir.child("temp").createDirectory()
         else:
+            if filesdir is not None:
+                raise ValueError("Only one of dbdir and filesdir"
+                                 " may be specified")
             if not isinstance(dbdir, filepath.FilePath):
                 dbdir = filepath.FilePath(dbdir)
                 # required subdirs: files, temp, run
@@ -1122,10 +1157,10 @@ class Store(Empowered):
         self.transact(self._startup)
 
         # _startup may have found some things which we must now upgrade.
-        if self._oldTypesRemaining:
+        if self._upgradeManager.upgradesPending:
             # Automatically upgrade when possible.
             self._upgradeComplete = PendingEvent()
-            d = self._upgradeService.addIterator(self._upgradeEverything())
+            d = self._upgradeService.addIterator(self._upgradeManager.upgradeEverything())
             def logUpgradeFailure(aFailure):
                 log.err(aFailure, "upgrading %r failed" % (self,))
                 return aFailure
@@ -1137,6 +1172,10 @@ class Store(Empowered):
         else:
             self._upgradeComplete = None
 
+        log.msg(
+            interface=iaxiom.IStatEvent,
+            store_opened=self.dbdir is not None and self.dbdir.path or '')
+
     _childCounter = 0
 
     def _attachChild(self, child):
@@ -1182,70 +1221,65 @@ class Store(Empowered):
         version information for upgrader service to run later.
         """
         typesToCheck = []
+
         for oid, module, typename, version in self.querySchemaSQL(_schema.ALL_TYPES):
             if self.debug:
                 print
                 print 'SCHEMA:', oid, module, typename, version
-            self.typenameAndVersionToID[typename, version] = oid
             if typename not in _typeNameToMostRecentClass:
                 try:
                     namedAny(module)
                 except ValueError, err:
                     raise ImportError('cannot find module ' + module, str(err))
+            self.typenameAndVersionToID[typename, version] = oid
 
+        # Can't call this until typenameAndVersionToID is populated, since this
+        # depends on building a reverse map of that.
+        persistedSchema = self._loadTypeSchema()
+
+        # Now that we have persistedSchema, loop over everything again and
+        # prepare old types.
+        for (typename, version), typeID in self.typenameAndVersionToID.iteritems():
             cls = _typeNameToMostRecentClass.get(typename)
 
             if cls is not None:
                 if version != cls.schemaVersion:
                     typesToCheck.append(
-                        self._prepareOldVersionOf(oid, typename, version))
+                        self._prepareOldVersionOf(
+                            typename, version, persistedSchema))
                 else:
                     typesToCheck.append(cls)
 
         for cls in typesToCheck:
-            self.checkTypeSchemaConsistency(cls)
+            self._checkTypeSchemaConsistency(cls, persistedSchema)
 
         # Schema is consistent!  Now, if I forgot to create any indexes last
         # time I saw this table, do it now...
+        extantIndexes = self._loadExistingIndexes()
         for cls in typesToCheck:
-            self._createIndexesFor(cls)
+            self._createIndexesFor(cls, extantIndexes)
 
-        cantUpgradeErrors = []
-        for oldVersion in self._oldTypesRemaining:
-            # We have to be able to get from oldVersion.schemaVersion to
-            # the most recent type.
+        self._upgradeManager.checkUpgradePaths()
 
-            currentType = _typeNameToMostRecentClass.get(
-                oldVersion.typeName, None)
 
-            if currentType is None:
-                # There isn't a current version of this type; it's entirely
-                # legacy, will be upgraded by deleting and replacing with
-                # something else.
-                continue
-
-            typeInQuestion = oldVersion.typeName
-            upgver = oldVersion.schemaVersion
-
-            while upgver < currentType.schemaVersion:
-                # Do we have enough of the schema present to upgrade?
-                if ((typeInQuestion, upgver)
-                    not in upgrade._upgradeRegistry):
-                    cantUpgradeErrors.append(
-                        "No upgrader present for %s (%s) from %d to %d" % (
-                            typeInQuestion, qual(currentType), upgver,
-                            upgver + 1))
-
-                # Is there a type available for each upgrader version?
-                if upgver+1 != currentType.schemaVersion:
-                    if (typeInQuestion, upgver+1) not in _legacyTypes:
-                        cantUpgradeErrors.append(
-                            "Type schema required for upgrade missing:"
-                            " %s version %d" % (
-                                typeInQuestion, upgver+1))
-                upgver += 1
-        if cantUpgradeErrors:
-            raise errors.NoUpgradePathAvailable('\n    '.join(cantUpgradeErrors))
+    def _loadExistingIndexes(self):
+        """
+        Return a C{set} of the SQL indexes which already exist in the
+        underlying database.  It is important to load all of this information
+        at once (as opposed to using many CREATE INDEX IF NOT EXISTS statements
+        or many CREATE INDEX statements and handling the errors) to minimize
+        the cost of opening a store.  Loading all the indexes at once is much
+        faster than doing pretty much anything that involves doing something
+        once per required index.
+        """
+        # Totally SQLite-specific: look up what indexes exist already in
+        # sqlite_master so we can skip trying to create them (which can be
+        # really slow).
+        return set(
+            name
+            for (name,) in self.querySchemaSQL(
+                "SELECT name FROM *DATABASE*.sqlite_master "
+                "WHERE type = 'index'"))
 
 
     def _initdb(self, dbfname):
@@ -1322,114 +1356,156 @@ class Store(Empowered):
 
         @return: an L{AtomicFile}.
         """
-        assert self.dbdir is not None, "Cannot create files in in-memory Stores (yet)"
         assert len(path) > 0, "newFile requires a nonzero number of segments"
-        tmpname = self.dbdir.child('temp').child(str(tempCounter.next()) + ".tmp")
+        if self.dbdir is None:
+            if self.filesdir is None:
+                raise RuntimeError("This in-memory store has no file directory")
+            else:
+                tmpbase = self.filesdir
+        else:
+            tmpbase = self.dbdir
+        tmpname = tmpbase.child('temp').child(str(tempCounter.next()) + ".tmp")
         return AtomicFile(tmpname.path, self.newFilePath(*path))
 
     def newDirectory(self, *path):
-        assert self.dbdir is not None, "Cannot create directories in in-memory Stores (yet)"
         p = self.filesdir
         for subdir in path:
             p = p.child(subdir)
         return p
 
-    def checkTypeSchemaConsistency(self, actualType):
-        """
-        Called for all known types at database startup: make sure that what we know
-        (in memory) about this type is
 
+    def _loadTypeSchema(self):
+        """
+        Load all of the stored schema information for all types known by this
+        store.  It's important to load everything all at once (rather than
+        loading the schema for each type separately as it is needed) to keep
+        store opening fast.  A single query with many results is much faster
+        than many queries with a few results each.
+
+        @return: A dict with two-tuples of item type name and schema version as
+            keys and lists of five-tuples of attribute schema information for
+            that type.  The elements of the five-tuple are::
+
+              - a string giving the name of the Python attribute
+              - a string giving the SQL type
+              - a boolean indicating whether the attribute is indexed
+              - the Python attribute type object (eg, axiom.attributes.integer)
+              - a string giving documentation for the attribute
+        """
+
+        # Oops, need an index going the other way.  This only happens once per
+        # store open, and it's based on data queried from the store, so there
+        # doesn't seem to be any broader way to cache and re-use the result.
+        # However, if we keyed the resulting dict on the database typeID rather
+        # than (typeName, schemaVersion), we wouldn't need the information this
+        # dict gives us.  That would mean changing the callers of this function
+        # to use typeID instead of that tuple, which may be possible.  Probably
+        # only represents a very tiny possible speedup.
+        typeIDToNameAndVersion = {}
+        for key, value in self.typenameAndVersionToID.iteritems():
+            typeIDToNameAndVersion[value] = key
+
+        # Indexing attribute, ordering by it, and getting rid of row_offset
+        # from the schema and the sorted() here doesn't seem to be any faster
+        # than doing this.
+        persistedSchema = sorted(self.querySchemaSQL(
+            "SELECT attribute, type_id, sqltype, indexed, "
+            "pythontype, docstring FROM *DATABASE*.axiom_attributes "))
+
+        # This is trivially (but measurably!) faster than getattr(attributes,
+        # pythontype).
+        getAttribute = attributes.__dict__.__getitem__
+
+        result = {}
+        for (attribute, typeID, sqltype, indexed, pythontype,
+             docstring) in persistedSchema:
+            key = typeIDToNameAndVersion[typeID]
+            if key not in result:
+                result[key] = []
+            result[key].append((
+                    attribute, sqltype, indexed,
+                    getAttribute(pythontype), docstring))
+        return result
+
+
+    def _checkTypeSchemaConsistency(self, actualType, onDiskSchema):
+        """
+        Called for all known types at database startup: make sure that what we
+        know (in memory) about this type agrees with what is stored about this
+        type in the database.
+
+        @param actualType: A L{MetaItem} instance which is associated with a
+            table in this store.  The schema it defines in memory will be
+            checked against the schema known in the database to ensure they
+            agree.
+
+        @param onDiskSchema: A mapping from L{MetaItem} instances (such as
+            C{actualType}) to the schema known in the database and associated
+            with C{actualType}.
+
+        @raise RuntimeError: if the schema defined by C{actualType} does not
+            match the database-present schema given in C{onDiskSchema} or if
+            C{onDiskSchema} contains a newer version of the schema associated
+            with C{actualType} than C{actualType} represents.
         """
         # make sure that both the runtime and the database both know about this
         # type; if they don't both know, we can't check that their views are
         # consistent
+        try:
+            inMemorySchema = _inMemorySchemaCache[actualType]
+        except KeyError:
+            inMemorySchema = _inMemorySchemaCache[actualType] = [
+                (storedAttribute.attrname, storedAttribute.sqltype)
+                for (name, storedAttribute) in actualType.getSchema()]
 
-        inMemorySchema = [(#storedAttribute.indexed,
-                           storedAttribute.sqltype,
-                           #storedAttribute.allowNone,
-                           storedAttribute.attrname)
-                          for (name, storedAttribute) in actualType.getSchema()]
-
-        # getTypeID is the wrong thing to do here because it's recursive!
-        typeID = self.typenameAndVersionToID[actualType.typeName,
-                                             actualType.schemaVersion]
-
-        onDiskSchema = [(ondisksqltype, ondiskattrname) for
-                        (ondiskindexed,
-                         ondisksqltype,
-                         ondiskallownone,
-                         ondiskattrname) in
-                        self.querySchemaSQL(_schema.IDENTIFYING_SCHEMA,
-                                           [typeID])]
-
-        if inMemorySchema != onDiskSchema:
+        key = (actualType.typeName, actualType.schemaVersion)
+        persistedSchema = [(storedAttribute[0], storedAttribute[1])
+                           for storedAttribute in onDiskSchema[key]]
+        if inMemorySchema != persistedSchema:
             raise RuntimeError(
                 "Schema mismatch on already-loaded %r <%r> object version %d: %r != %r" %
                 (actualType, actualType.typeName, actualType.schemaVersion,
                  onDiskSchema, inMemorySchema))
 
-
         if actualType.__legacy__:
             return
 
-        if self.querySchemaSQL(_schema.GET_GREATER_VERSIONS_OF_TYPE,
-                               [actualType.typeName,
-                                actualType.schemaVersion]):
+        if (key[0], key[1] + 1) in onDiskSchema:
             raise RuntimeError(
                 "Greater versions of database %r objects in the DB than in memory" %
                 (actualType.typeName,))
 
-        # finally find old versions of the data and prepare to upgrade it.
 
-    def _prepareOldVersionOf(self, typeID, typename, version):
+    # finally find old versions of the data and prepare to upgrade it.
+    def _prepareOldVersionOf(self, typename, version, persistedSchema):
         """
         Note that this database contains old versions of a particular type.
-        Create the appropriate dummy item subclass.
-        """
+        Create the appropriate dummy item subclass and queue the type to be
+        upgraded.
 
-        appropriateSchema = self.querySchemaSQL(_schema.SCHEMA_FOR_TYPE, [typeID])
+        @param typename: The I{typeName} associated with the schema for which
+            to create a dummy item class.
+
+        @param version: The I{schemaVersion} of the old version of the schema
+            for which to create a dummy item class.
+
+        @param persistedSchema: A mapping giving information about all schemas
+            stored in the database, used to create the attributes of the dummy
+            item class.
+        """
+        appropriateSchema = persistedSchema[typename, version]
         # create actual attribute objects
         dummyAttributes = {}
-        for indexed, pythontype, attribute, docstring in appropriateSchema:
-            atr = getattr(attributes, pythontype)(indexed=indexed,
-                                                  doc=docstring)
+        for (attribute, sqlType, indexed, pythontype,
+             docstring) in appropriateSchema:
+            atr = pythontype(indexed=indexed, doc=docstring)
             dummyAttributes[attribute] = atr
         dummyBases = []
-        dis = declareLegacyItem(typename, version, dummyAttributes, dummyBases)
-        self._oldTypesRemaining.append(dis)
-        return dis
+        oldType = declareLegacyItem(
+            typename, version, dummyAttributes, dummyBases)
+        self._upgradeManager.queueTypeUpgrade(oldType)
+        return oldType
 
-    _anyUpgradesThisTypeYet = False
-
-    def _upgradeOneThing(self):
-        """
-        Upgrade one Item; return True if there may be more work to do, False if
-        this store is definitely fully upgraded.
-        """
-        while self._oldTypesRemaining:
-            t0 = self._oldTypesRemaining[0]
-            onething = list(self.query(t0, limit=1))
-            if not onething:
-                self._oldTypesRemaining.pop(0)
-                if self._anyUpgradesThisTypeYet:
-                    log.msg("%s finished upgrading %s" % (self.dbdir.path, qual(t0)))
-                self._anyUpgradesThisTypeYet = False
-                continue
-            o = onething[0]
-            self._anyUpgradesThisTypeYet = True
-            self.transact(upgrade.upgradeAllTheWay, o)
-            return True
-        return False
-
-    def _upgradeEverything(self):
-        didAny = False
-        while self._upgradeOneThing():
-            if not didAny:
-                didAny = True
-                log.msg("%s beginning upgrade..." % (self.dbdir.path,))
-            yield None
-        if didAny:
-            log.msg("%s completely upgraded." % (self.dbdir.path,))
 
     def whenFullyUpgraded(self):
         """
@@ -1626,6 +1702,12 @@ class Store(Empowered):
 
 
     def changed(self, item):
+        """
+        An item in this store was changed.  Add it to the current transaction's
+        list of changed items, if a transaction is currently underway, or raise
+        an exception if this L{Store} is currently in a state which does not
+        allow changes.
+        """
         if self._rejectChanges:
             raise errors.ChangeRejected()
         if self.transaction is not None:
@@ -1673,13 +1755,10 @@ class Store(Empowered):
             self._cleanupTxnState()
 
     # The following three methods are necessary...
-
     # - in PySQLite: because PySQLite has some buggy transaction handling which
     #   makes it impossible to issue explicit BEGIN statements - which we
     #   _need_ to do to provide guarantees for read/write transactions.
 
-    # - in APSW: because there are no .commit() or .rollback() methods.
-
     def _begin(self):
         if self.debug:
             print '<'*10, 'BEGIN', '>'*10
@@ -1785,14 +1864,29 @@ class Store(Empowered):
                                         int( (sum(l)/len(l)) * 1000000.),)
 
     def _indexNameOf(self, tableClass, attrname):
-        return "%s.axiomidx_%s_v%d_%s" % (self.databaseName,
-                                          tableClass.typeName,
-                                          tableClass.schemaVersion,
-                                          '_'.join(attrname))
+        """
+        Return the unqualified (ie, no database name) name of the given
+        attribute of the given table.
+
+        @type tableClass: L{MetaItem}
+        @param tableClass: The Python class associated with a table in the
+            database.
+
+        @param attrname: A sequence of the names of the columns of the
+            indicated table which will be included in the named index.
+
+        @return: A C{str} giving the name of the index which will index the
+            given attributes of the given table.
+        """
+        return "axiomidx_%s_v%d_%s" % (tableClass.typeName,
+                                       tableClass.schemaVersion,
+                                       '_'.join(attrname))
+
 
     def _tableNameFor(self, typename, version):
         return "%s.item_%s_v%d" % (self.databaseName, typename, version)
 
+
     def getTableName(self, tableClass):
         """
         Retrieve the fully qualified name of the table holding items
@@ -1932,7 +2026,12 @@ class Store(Empowered):
         if self.tablesCreatedThisTransaction is not None:
             self.tablesCreatedThisTransaction.append(tableClass)
 
-        self._createIndexesFor(tableClass)
+        # We can pass () for extantIndexes here because since the table didn't
+        # exist for tableClass, none of its indexes could have either.
+        # Whatever checks _createIndexesFor will make would give the same
+        # result against the actual set of existing indexes as they will
+        # against ().
+        self._createIndexesFor(tableClass, ())
 
         for n, (name, storedAttribute) in enumerate(tableClass.getSchema()):
             self.executeSchemaSQL(
@@ -1947,14 +2046,30 @@ class Store(Empowered):
         return typeID
 
 
-    def _createIndexesFor(self, tableClass):
-        indexes = set()
-        for nam, atr in tableClass.getSchema():
-            if atr.indexed:
-                indexes.add(((atr.getShortColumnName(self),), (atr.attrname,)))
-            for compound in atr.compoundIndexes:
-                indexes.add((tuple(inatr.getShortColumnName(self) for inatr in compound),
-                             tuple(inatr.attrname for inatr in compound)))
+    def _createIndexesFor(self, tableClass, extantIndexes):
+        """
+        Create any indexes which don't exist and are required by the schema
+        defined by C{tableClass}.
+
+        @param tableClass: A L{MetaItem} instance which may define a schema
+            which includes indexes.
+
+        @param extantIndexes: A container (anything which can be the right-hand
+            argument to the C{in} operator) which contains the unqualified
+            names of all indexes which already exist in the underlying database
+            and do not need to be created.
+        """
+        try:
+            indexes = _requiredTableIndexes[tableClass]
+        except KeyError:
+            indexes = set()
+            for nam, atr in tableClass.getSchema():
+                if atr.indexed:
+                    indexes.add(((atr.getShortColumnName(self),), (atr.attrname,)))
+                for compound in atr.compoundIndexes:
+                    indexes.add((tuple(inatr.getShortColumnName(self) for inatr in compound),
+                                 tuple(inatr.attrname for inatr in compound)))
+            _requiredTableIndexes[tableClass] = indexes
 
         # _ZOMFG_ SQL is such a piece of _shit_: you can't fully qualify the
         # table name in CREATE INDEX statements because the _INDEX_ is fully
@@ -1963,16 +2078,13 @@ class Store(Empowered):
         indexColumnPrefix = '.'.join(self.getTableName(tableClass).split(".")[1:])
 
         for (indexColumns, indexAttrs) in indexes:
-            csql = ('CREATE INDEX %s ON %s(%s)' %
-                    (self._indexNameOf(tableClass, indexAttrs),
-                     indexColumnPrefix,
-                     ', '.join(indexColumns)))
-            try:
-                self.createSQL(csql)
-            except errors.SQLError, sqle:
-                # Ignore duplicate indexes.
-                if "already exists" not in str(sqle):
-                    raise
+            nameOfIndex = self._indexNameOf(tableClass, indexAttrs)
+            if nameOfIndex in extantIndexes:
+                continue
+            csql = 'CREATE INDEX %s.%s ON %s(%s)' % (
+                self.databaseName, nameOfIndex, indexColumnPrefix,
+                ', '.join(indexColumns))
+            self.createSQL(csql)
 
 
     def getTableQuery(self, typename, version):
@@ -2017,7 +2129,7 @@ class Store(Empowered):
         if not isinstance(storeID, (int, long)):
             raise TypeError("storeID *must* be an int or long, not %r" % (
                     type(storeID).__name__,))
-        if storeID == -1:
+        if storeID == STORE_SELF_ID:
             return self
         if self.objectCache.has(storeID):
             return self.objectCache.get(storeID)
@@ -2075,7 +2187,7 @@ class Store(Empowered):
                 # upgradeVersion will do caching as necessary, we don't have to
                 # cache here.  (It must, so that app code can safely call
                 # upgradeVersion and get a consistent object out of it.)
-                x = self.transact(upgrade.upgradeAllTheWay, x)
+                x = self.transact(self._upgradeManager.upgradeItem, x)
             elif not x.__legacy__:
                 # We loaded the most recent version of an object
                 self.objectCache.cache(storeID, x)
