@@ -6,6 +6,7 @@ from twisted.trial.unittest import TestCase, SkipTest
 from axiom.iaxiom import IComparison, IColumn
 from axiom.store import Store, ItemQuery
 from axiom.item import Item, Placeholder
+from axiom.test.util import QueryCounter
 
 from axiom import errors
 from axiom.attributes import (
@@ -353,6 +354,29 @@ class BasicQuery(TestCase):
         s.close()
 
 
+    def test_distinctQuerySQLiteBug(self):
+        """
+        Test for an SQLite bug.
+
+        SQLite versions 3.5.8 and 3.5.9 have a bug that causes incorrect query
+        results under certain circumstances:
+
+            - A column with an index
+            - SELECT DISTINCT query...
+            - ... with an IS NOT NULL comparison on that column.
+
+        Upstream ticket: http://www.sqlite.org/cvstrac/tktview?tn=3236
+        """
+        s = Store()
+        def entesten():
+            C(store=s, name=u'a')
+            C(store=s, name=u'b')
+            q = s.query(C, C.name != None).getColumn('name').distinct()
+            self.assertEqual(sorted(q), [u'a', u'b'])
+        s.transact(entesten)
+        s.close()
+
+
     def test_itemQueryDistinct(self):
         """
         Verify that a join which would produce duplicate rows will not produce
@@ -1012,36 +1036,12 @@ class QueryComplexity(TestCase):
     def setUp(self):
         self.store = Store()
         self.query = self.store.query(FirstType, self.comparison)
-        class counter:
-            count = 0
-            def increment(self):
-                self.count += 1
-                return 0
-        self.counter = counter()
-
-        store = self.store
-        connection = store.connection
-        _connection = connection._connection
-        try:
-            self.setProgressHandler = _connection.set_progress_handler
-        except AttributeError:
-            raise SkipTest(
-                "Axiom missing setProgressHandler, cannot run bytecode "
-                "execution-counting tests.")
 
         # Make one of each to get any initialization taken care of so it
         # doesn't pollute our numbers below.
         FirstType(store=self.store)
         SecondType(store=self.store)
 
-        self.setProgressHandler(self.counter.increment)
-
-
-    def _countQuery(self):
-        for i in range(10):
-            self.counter.count = 0
-            list(self.query)
-            yield self.counter.count
 
     def test_firstTableOuterLoop(self):
         """
@@ -1052,9 +1052,10 @@ class QueryComplexity(TestCase):
         Test this by inserting rows into the first table and checking that the
         number of bytecodes executed increased.
         """
+        counter = QueryCounter(self.store)
         counts = []
-        for c in self._countQuery():
-            counts.append(c)
+        for c in range(10):
+            counts.append(counter.measure(list, self.query))
             FirstType(store=self.store)
 
         # Make sure they're not all the same
@@ -1074,8 +1075,10 @@ class QueryComplexity(TestCase):
         condition.  This should mean that rows from the second table are
         never examined.
         """
+        counter = QueryCounter(self.store)
         count = None
-        for c in self._countQuery():
+        for i in range(10):
+            c = counter.measure(list, self.query)
             if count is None:
                 count = c
             self.assertEqual(count, c)
