@@ -8,6 +8,7 @@ from twisted.trial.unittest import TestCase
 from twisted.application.service import IService
 from twisted.internet.defer import Deferred
 from twisted.internet.task import Clock
+from twisted.python import filepath
 
 from epsilon.extime import Time
 
@@ -25,21 +26,16 @@ class TestEvent(Item):
     typeName = 'test_event'
     schemaVersion = 1
 
-    deferred = inmemory()       # these won't fall out of memory due to
+    testCase = inmemory()       # this won't fall out of memory due to
                                 # caching, thanks.
-    testCase = inmemory()
-
     name = text()
 
-    maxRunCount = integer()     # fail the test if we run more than this many
-                                # times
     runCount = integer()
     runAgain = integer()        # milliseconds to add, then run again
     winner = integer(default=0) # is this the event that is supposed to
                                 # complete the test successfully?
 
     def __init__(self, **kw):
-        self.deferred = None
         super(TestEvent, self).__init__(**kw)
         self.runCount = 0
 
@@ -55,22 +51,18 @@ class TestEvent(Item):
                 if isinstance(evt.runnable, _SubSchedulerParentHook):
                     count += 1
             if count > 1:
-                return self.fail("Too many TimedEvents for the SubStore", count)
+                return self.fail("Too many TimedEvents for the SubStore: %d" % (count,))
 
         self.runCount += 1
-        if self.runCount > self.maxRunCount:
-            return self.fail("%s ran too many times"% (self.name))
         if self.runAgain is not None:
-            result = Time() + timedelta(milliseconds=self.runAgain)
+            result = self.testCase.now() + timedelta(seconds=self.runAgain)
             self.runAgain = None
         else:
-            if self.winner and self.deferred is not None:
-                self.deferred.callback('done')
             result = None
         return result
 
-    def fail(self, *msg):
-        self.deferred.errback(self.testCase.failureException(*msg))
+    def fail(self, msg):
+        self.testCase.fail(msg)
 
 
 class NotActuallyRunnable(Item):
@@ -92,11 +84,43 @@ class SpecialErrorHandler(Item):
         failureObj.trap(SpecialError)
         self.procd = 1
 
+
+
+class HookRunner(Item):
+    """
+    Runnable that simply calls a supplied hook.
+    """
+    ignored = integer()
+    hook = inmemory()
+
+    def run(self):
+        self.hook(self)
+
+
+
 class SchedTest:
     def tearDown(self):
         return IService(self.siteStore).stopService()
 
 
+    def setUp(self):
+        self.clock = Clock()
+
+        scheduler = Scheduler(store=self.siteStore)
+        self.stubTime(scheduler)
+        installOn(scheduler, self.siteStore)
+        IService(self.siteStore).startService()
+
+
+    def now(self):
+        return Time.fromPOSIXTimestamp(self.clock.seconds())
+
+
+    def stubTime(self, scheduler):
+        scheduler.callLater = self.clock.callLater
+        scheduler.now = self.now
+
+
     def test_implementsSchedulerInterface(self):
         """
         Verify that IScheduler is declared as implemented.
@@ -114,52 +138,48 @@ class SchedTest:
         # only once, and that the third never fires.
         s = self.store
 
-        d = Deferred()
-
-        interval = 30
-
         t1 = TestEvent(testCase=self,
                        name=u't1',
-                       store=s, maxRunCount=1, runAgain=None, deferred=d)
+                       store=s, runAgain=None)
         t2 = TestEvent(testCase=self,
                        name=u't2',
-                       store=s, maxRunCount=2, runAgain=interval, deferred=d, winner=True)
+                       store=s, runAgain=2)
         t3 = TestEvent(testCase=self,
                        name=u't3',
-                       store=s, maxRunCount=0, runAgain=None, deferred=d)
+                       store=s, runAgain=None)
 
-        now = Time()
+        now = self.now()
         self.ts = [t1, t2, t3]
 
         S = IScheduler(s)
 
         # Schedule them out of order to make sure behavior doesn't
         # depend on tasks arriving in soonest-to-latest order.
-        S.schedule(t2, now + timedelta(milliseconds=interval * 2))
-        S.schedule(t1, now + timedelta(milliseconds=interval * 1))
-        S.schedule(t3, now + timedelta(milliseconds=interval * 100))
+        S.schedule(t2, now + timedelta(seconds=3))
+        S.schedule(t1, now + timedelta(seconds=1))
+        S.schedule(t3, now + timedelta(seconds=100))
 
-        return d
+        self.clock.pump([2, 2, 2])
+        self.assertEqual(t1.runCount, 1)
+        self.assertEqual(t2.runCount, 2)
+        self.assertEqual(t3.runCount, 0)
 
 
     def test_unscheduling(self):
         """
         Test the unscheduleFirst method of the scheduler.
         """
-        now = Time()
         d = Deferred()
         sch = IScheduler(self.store)
-        t1 = TestEvent(testCase=self, name=u't1', store=self.store, maxRunCount=0)
-        t2 = TestEvent(testCase=self, name=u't2', store=self.store, maxRunCount=1, runAgain=None, winner=True, deferred=d)
+        t1 = TestEvent(testCase=self, name=u't1', store=self.store)
+        t2 = TestEvent(testCase=self, name=u't2', store=self.store, runAgain=None)
 
-        # Make sure the inmemory attributes hang around
-        self.ts = [t1, t2]
-
-        sch.schedule(t1, now + timedelta(milliseconds=100))
-        sch.schedule(t2, now + timedelta(milliseconds=200))
+        sch.schedule(t1, self.now() + timedelta(seconds=1))
+        sch.schedule(t2, self.now() + timedelta(seconds=2))
         sch.unscheduleFirst(t1)
-
-        return d
+        self.clock.advance(3)
+        self.assertEquals(t1.runCount, 0)
+        self.assertEquals(t2.runCount, 1)
 
 
     def test_inspection(self):
@@ -167,7 +187,7 @@ class SchedTest:
         Test that the L{scheduledTimes} method returns an iterable of all the
         times at which a particular item is scheduled to run.
         """
-        now = Time() + timedelta(seconds=1)
+        now = self.now() + timedelta(seconds=1)
         off = timedelta(seconds=3)
         sch = IScheduler(self.store)
         runnable = TestEvent(store=self.store, name=u'Only event')
@@ -180,6 +200,24 @@ class SchedTest:
             [now, now + off, now + off + off])
 
 
+    def test_scheduledTimesDuringRun(self):
+        """
+        L{Scheduler.scheduledTimes} should not include scheduled times that have
+        already triggered.
+        """
+        futureTimes = []
+        scheduler = IScheduler(self.store)
+        runner = HookRunner(
+            store=self.store,
+            hook=lambda self: futureTimes.append(
+                list(scheduler.scheduledTimes(self))))
+
+        then = self.now() + timedelta(seconds=1)
+        scheduler.schedule(runner, self.now())
+        scheduler.schedule(runner, then)
+        self.clock.advance(1)
+        self.assertEquals(futureTimes, [[then], []])
+
 
     def test_deletedRunnable(self):
         """
@@ -187,7 +225,7 @@ class SchedTest:
         L{TimedEvent.invokeRunnable} just deletes the L{TimedEvent} without
         raising an exception.
         """
-        now = Time()
+        now = self.now()
         scheduler = IScheduler(self.store)
         runnable = TestEvent(store=self.store, name=u'Only event')
         scheduler.schedule(runnable, now)
@@ -210,52 +248,49 @@ class SchedTest:
 
 class TopStoreSchedTest(SchedTest, TestCase):
     def setUp(self):
-        # self.storePath = self.mktemp()
         self.store = self.siteStore = Store()
-        installOn(Scheduler(store=self.store), self.store)
-        IService(self.store).startService()
+        super(TopStoreSchedTest, self).setUp()
 
 
     def testBasicScheduledError(self):
         S = IScheduler(self.store)
-        now = Time()
-        S.schedule(NotActuallyRunnable(store=self.store), now)
-        d = Deferred()
+        S.schedule(NotActuallyRunnable(store=self.store), self.now())
+
         te = TestEvent(store=self.store, testCase=self,
-                       name=u't1', maxRunCount=1, runAgain=None,
-                       winner=True, deferred=d)
-        self.te = te            # don't gc the deferred
-        now2 = Time()
-        S.schedule(te, now2)
+                       name=u't1', runAgain=None)
+        S.schedule(te, self.now() + timedelta(seconds=1))
+
         self.assertEquals(
             self.store.query(TimedEventFailureLog).count(), 0)
-        def later(result):
-            errs = self.flushLoggedErrors(AttributeError)
-            self.assertEquals(len(errs), 1)
-            self.assertEquals(self.store.query(TimedEventFailureLog).count(), 1)
-        return d.addCallback(later)
+
+        self.clock.advance(3)
+
+        self.assertEquals(te.runCount, 1)
+
+        errs = self.flushLoggedErrors(AttributeError)
+        self.assertEquals(len(errs), 1)
+        self.assertEquals(self.store.query(TimedEventFailureLog).count(), 1)
 
     def testScheduledErrorWithHandler(self):
         S = IScheduler(self.store)
-        now = Time()
         spec = SpecialErrorHandler(store=self.store)
-        S.schedule(spec, now)
-        d = Deferred()
+        S.schedule(spec, self.now())
+
         te = TestEvent(store=self.store, testCase=self,
-                       name=u't1', maxRunCount=1, runAgain=None,
-                       winner=True, deferred=d)
-        self.te = te            # don't gc the deferred
-        now2 = Time()
-        S.schedule(te, now2)
+                       name=u't1', runAgain=None)
+        S.schedule(te, self.now() + timedelta(seconds=1))
         self.assertEquals(
             self.store.query(TimedEventFailureLog).count(), 0)
-        def later(result):
-            errs = self.flushLoggedErrors(SpecialError)
-            self.assertEquals(len(errs), 1)
-            self.assertEquals(self.store.query(TimedEventFailureLog).count(), 0)
-            self.failUnless(spec.procd)
-            self.failIf(spec.broken)
-        return d.addCallback(later)
+
+        self.clock.advance(3)
+
+        self.assertEquals(te.runCount, 1)
+
+        errs = self.flushLoggedErrors(SpecialError)
+        self.assertEquals(len(errs), 1)
+        self.assertEquals(self.store.query(TimedEventFailureLog).count(), 0)
+        self.failUnless(spec.procd)
+        self.failIf(spec.broken)
 
 
 
@@ -268,18 +303,52 @@ class SubSchedulerTests(SchedTest, TestCase):
         Create a site store for the substore which will contain the IScheduler
         being tested.  Start its IService so any scheduled events will run.
         """
-        self.storePath = self.mktemp()
+        self.storePath = filepath.FilePath(self.mktemp())
         self.siteStore = Store(self.storePath)
-        self.svc = IService(self.siteStore)
-        self.svc.startService()
+        super(SubSchedulerTests, self).setUp()
 
         substoreItem = SubStore.createNew(self.siteStore, ['scheduler_test'])
         self.substore = substoreItem.open()
-        installOn(SubScheduler(store=self.substore), self.substore)
+        self.scheduler = scheduler = SubScheduler(store=self.substore)
+        self.stubTime(scheduler)
+        installOn(scheduler, self.substore)
 
         self.store = self.substore
 
 
+    def test_scheduleAfterParentHookError(self):
+        """
+        A transient error during a L{_SubSchedulerParentHook} run (such as
+        failing to open the substore for whatever reason) should not disable
+        subsequent scheduling.
+        """
+        runnable = TestEvent(store=self.store)
+
+        # Schedule runnable, but fail the _SubSchedulerParentHook run.
+
+        self.scheduler.schedule(runnable, self.now() + timedelta(seconds=1))
+        hook = self.siteStore.findUnique(_SubSchedulerParentHook)
+        def stumble():
+            raise IOError('Denied')
+        object.__setattr__(hook, 'run', stumble)
+        self.clock.advance(1)
+        object.__delattr__(hook, 'run')
+
+        self.assertEquals(
+            self.siteStore.findUnique(TimedEventFailureLog).runnable,
+            hook)
+        [err] = self.flushLoggedErrors(IOError)
+        self.assertEquals(str(err.value), 'Denied')
+        self.assertEquals(runnable.runCount, 0)
+
+        # Schedule runnable again.  The restored hook in the site store should
+        # trigger both scheduled runs in the substore now.
+
+        self.scheduler.schedule(runnable, self.now() + timedelta(seconds=1))
+        self.clock.advance(1)
+        self.assertEquals(runnable.runCount, 2)
+
+
 
 class SchedulerStartupTests(TestCase):
     """
@@ -413,7 +482,7 @@ class MissingService(unittest.TestCase):
         controlled by these tests.
         """
         self.calls = []
-        self.store = Store(self.mktemp())
+        self.store = Store(filepath.FilePath(self.mktemp()))
         self.siteScheduler = Scheduler(store=self.store)
         installOn(self.siteScheduler, self.store)
         self.siteScheduler.callLater = self._callLater
@@ -494,7 +563,7 @@ class SubStoreSchedulerReentrancy(TestCase):
     def setUp(self):
         self.clock = Clock()
 
-        self.dbdir = self.mktemp()
+        self.dbdir = filepath.FilePath(self.mktemp())
         self.store = Store(self.dbdir)
         self.substoreItem = SubStore.createNew(self.store, ['sub'])
         self.substore = self.substoreItem.open()
