@@ -1,7 +1,39 @@
+# Copright 2008 Divmod, Inc.  See LICENSE file for details.
 # -*- test-case-name: axiom.test.test_userbase -*-
 
+"""
+The L{axiom.userbase} module implements various interfaces from L{twisted.cred}
+to allow an Axiom database to serve as an integration point for Twisted
+services that do authentication.
+
+While not strictly required, one part of this implementation is the idiom that
+Axiom (by default) partitions its user database into a separate data-store for
+each users.
+
+This has several advantages:
+
+  - Each user's account can be quickly and independently added to or removed
+    from the system; inactive accounts can be quickly moved to archival
+    storage.
+
+  - User accounts may be migrated between servers relatively easily.
+
+  - Database queries that deal with a single user's data are completely
+    partitioned; even naive and inefficient queries can still be run quickly as
+    long as users do not individually have a lot of data in a particular table.
+
+For truly multi-user applications, this partitioning is incomplete without an
+abstract facility for exchanging data between different users of the same
+application.  This module does not implement such a facility, as it is left to
+higher-level mechanisms such as Mantissa's messaging system in
+L{xmantissa.messaging}.  However, this module works standalone as well; just be
+aware that a user's database contains only their own data.
+"""
+
 import warnings
 
+from zope.interface import implements, Interface
+
 from twisted.cred.portal import IRealm
 from twisted.cred.credentials import IUsernamePassword, IUsernameHashedPassword
 from twisted.cred.checkers import ICredentialsChecker, ANONYMOUS
@@ -10,18 +42,19 @@ from twisted.python import log
 from axiom.store import Store
 from axiom.item import Item
 from axiom.substore import SubStore
-from axiom.attributes import text, integer, reference, boolean, AND
-from axiom.errors import BadCredentials, NoSuchUser, DuplicateUser
+from axiom.attributes import text, integer, reference, boolean, AND, OR
+from axiom.errors import (
+    BadCredentials, NoSuchUser, DuplicateUser, MissingDomainPart)
+from axiom.scheduler import IScheduler
 from axiom import upgrade, iaxiom
 
-from zope.interface import implements, Interface, Attribute
-
 ANY_PROTOCOL = u'*'
 
 def dflip(x):
     warnings.warn("Don't use dflip no more", stacklevel=2)
     return x
 
+
 class AllNamesConflict(Exception):
     """
     When inserting a SubStore into a site store, no names were found which were
@@ -31,6 +64,7 @@ class AllNamesConflict(Exception):
     and the site database is not modified.
     """
 
+
 class DatabaseDirectoryConflict(Exception):
     """
     When inserting a SubStore into a site store, the selected ultimate location
@@ -40,26 +74,43 @@ class DatabaseDirectoryConflict(Exception):
     and the site database is not modified.
     """
 
+
+
 class IPreauthCredentials(Interface):
     """
-    Credentials used to indicate the user has indeed authenticated
-    already, even though we have no credentials to present at the moment.
+    Deprecated.  Don't use this.  If you wrote a checker which can check this
+    interface, make it check one of the interfaces L{Preauthenticated}
+    implements, instead.
     """
-    username = Attribute("username@domain-style string")
+
 
 
 class Preauthenticated(object):
-    implements(IPreauthCredentials)
+    """
+    A credentials object of multiple types which has already been authenticated
+    somehow.
+
+    Credentials interfaces methods are implemented to behave as if the correct
+    credentials had been supplied.
+    """
+    implements(IUsernamePassword, IUsernameHashedPassword)
 
     def __init__(self, username):
         self.username = username
 
+
     def checkPassword(self, password):
-        # XXX should we really be implementing this here?  hmm.  sip tests
-        # require it but it seems wrong.
+        """
+        The password checks out.
+        """
         return True
 
 
+    def __repr__(self):
+        return '<Preauthenticated: %s>' % (self.username,)
+
+
+
 class LoginMethod(Item):
     typeName = 'login_method'
     schemaVersion = 2
@@ -145,6 +196,9 @@ class LoginAccount(Item):
             for x in list(oldAccounts) + list(oldMethods):
                 x.deleteFromStore()
             self.cloneInto(ss, ss)
+            sched = IScheduler(ss, None)
+            if sched is not None:
+                sched.migrateDown()
         ss.transact(_)
 
     def migrateUp(self):
@@ -160,6 +214,9 @@ class LoginAccount(Item):
             # you're making a mistake. -glyph
             siteStoreSubRef = siteStore.getItemByID(self.store.idInParent)
             self.cloneInto(siteStore, siteStoreSubRef)
+            sched = IScheduler(self.store, None)
+            if sched is not None:
+                sched.migrateUp()
         siteStore.transact(_)
 
     def cloneInto(self, newStore, avatars):
@@ -174,13 +231,13 @@ class LoginAccount(Item):
                           disabled=self.disabled)
         for siteMethod in self.store.query(LoginMethod,
                                            LoginMethod.account == self):
-            lm = LoginMethod(store=newStore,
-                             localpart=siteMethod.localpart,
-                             domain=siteMethod.domain,
-                             internal=siteMethod.internal,
-                             protocol=siteMethod.protocol,
-                             verified=siteMethod.verified,
-                             account=la)
+            LoginMethod(store=newStore,
+                        localpart=siteMethod.localpart,
+                        domain=siteMethod.domain,
+                        internal=siteMethod.internal,
+                        protocol=siteMethod.protocol,
+                        verified=siteMethod.verified,
+                        account=la)
         return la
 
     def deleteLoginMethods(self):
@@ -359,11 +416,9 @@ class LoginBase:
     """
     implements(IRealm, ICredentialsChecker)
 
-    credentialInterfaces = (IUsernamePassword, IUsernameHashedPassword, IPreauthCredentials)
+    credentialInterfaces = (IUsernamePassword, IUsernameHashedPassword)
 
-    def installOn(self, other):
-        other.powerUp(self, IRealm)
-        other.powerUp(self, ICredentialsChecker)
+    powerupInterfaces = (IRealm, ICredentialsChecker)
 
     def accountByAddress(self, username, domain):
         """
@@ -465,28 +520,27 @@ class LoginBase:
         raise NotImplementedError()
 
     def requestAvatarId(self, credentials):
-        passwordSecure = IUsernameHashedPassword(credentials, None) is not None
-        # ^ need to do something with this.  security warning perhaps?
-
         try:
             username, domain = credentials.username.split('@', 1)
-            username = unicode(username)
-            domain = unicode(domain)
-
-            acct = self.accountByAddress(username, domain)
-            if acct is not None:
-                if IPreauthCredentials.providedBy(credentials):
-                    return acct.storeID
-                else:
-                    password = acct.password
-                    if credentials.checkPassword(password):
-                        return acct.storeID
-                    else:
-                        raise BadCredentials()
-            raise NoSuchUser(credentials.username)
-        except:
+        except ValueError:
             self.failedLogins += 1
-            raise
+            raise MissingDomainPart(credentials.username)
+
+        username = unicode(username)
+        domain = unicode(domain)
+
+        acct = self.accountByAddress(username, domain)
+        if acct is not None:
+            password = acct.password
+            if credentials.checkPassword(password):
+                return acct.storeID
+            else:
+                self.failedLogins += 1
+                raise BadCredentials()
+
+        self.failedLogins += 1
+        raise NoSuchUser(credentials.username)
+
 
 class LoginSystem(Item, LoginBase, SubStoreLoginMixin):
     schemaVersion = 1
@@ -496,6 +550,18 @@ class LoginSystem(Item, LoginBase, SubStoreLoginMixin):
     failedLogins = integer(default=0)
 
 
+def getLoginMethods(store, protocol=None):
+    """
+    Retrieve L{LoginMethod} items from store C{store}, optionally constraining
+    them by protocol
+    """
+    if protocol is not None:
+        comp = OR(LoginMethod.protocol == u'*',
+                  LoginMethod.protocol == protocol)
+    else:
+        comp = None
+    return store.query(LoginMethod, comp)
+
 def getAccountNames(store, protocol=None):
     """
     Retrieve account name information about the given database.
@@ -507,13 +573,8 @@ def getAccountNames(store, protocol=None):
     @return: A generator of two-tuples of (username, domain) which
     refer to the given store.
     """
-    if protocol:
-        for meth in store.query(LoginMethod,
-                                LoginMethod.protocol == protocol):
-            yield (meth.localpart, meth.domain)
-    else:
-        for meth in store.query(LoginMethod):
-            yield (meth.localpart, meth.domain)
+    return ((meth.localpart, meth.domain) for meth
+                in getLoginMethods(store, protocol))
 
 
 def getDomainNames(store):
